from pathlib import Path

from pydantic import BaseModel
from loguru import logger

from .metadata import Metadata


class Image(BaseModel):
    """
    A single image object in the return of `generate_image` method.
    """

    filename: str
    data: bytes
    metadata: Metadata

    def __str__(self):
        return f"Image(filename={self.filename})"

    __repr__ = __str__

    def save(
        self, path: str = "temp", filename: str | None = None, verbose: bool = False
    ):
        """
        Save image to local file.

        Parameters
        ----------
        path : `str`, optional
            Path to save the image, by default will save to ./temp
        filename : `str`, optional
            Filename of the saved file, by default will use `self.filename`
            If provided, `self.filename` will also be updated to match this value
        verbose : `bool`, optional
            If True, print the path of the saved file, by default False
        """
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)

        self.filename = filename or self.filename
        dest = path / self.filename
        dest.write_bytes(self.data)

        if verbose:
            logger.info(f"Image saved as {dest.resolve()}")



from pydantic import BaseModel


class User(BaseModel):
    """
    NovelAI credentials for user account authentication.
    """

    username: str
    password: str

    def __str__(self):
        return f"User(username={self.username})"

    __repr__ = __str__



from enum import Enum

from .types import HostInstance


HEADERS = {
    "Content-Type": "application/json",
    "Origin": "https://novelai.net",
    "Referer": "https://novelai.net",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
}


class Host(Enum):
    API = HostInstance(
        url="https://api.novelai.net", accept="application/x-zip-compressed"
    )
    WEB = HostInstance(url="https://image.novelai.net", accept="binary/octet-stream")


class Endpoint(Enum):
    LOGIN = "/user/login"
    USERDATA = "/user/data"
    IMAGE = "/ai/generate-image"


class Model(Enum):
    # Anime V3
    V3 = "nai-diffusion-3"
    V3INP = "nai-diffusion-3-inpainting"

    # Furry model beta v1.3
    # Note that prompt preset in Metadata added by qualityToggle and ucPreset could be different,
    # but this module will not be specially adapted for it until a stable version is released.
    FURRY = "nai-diffusion-furry"
    FURRYINP = "furry-diffusion-inpainting"


class Controlnet(Enum):
    PALETTESWAP = "hed"
    FORMLOCK = "midas"
    SCRIBBLER = "fake_scribble"
    BUILDINGCONTROL = "mlsd"
    LANDSCAPER = "uniformer"


class Action(Enum):
    GENERATE = "generate"
    INPAINT = "infill"
    IMG2IMG = "img2img"


class Resolution(Enum):
    SMALL_PORTRAIT = (512, 768)
    SMALL_LANDSCAPE = (768, 512)
    SMALL_SQUARE = (640, 640)
    NORMAL_PORTRAIT = (832, 1216)
    NORMAL_LANDSCAPE = (1216, 832)
    NORMAL_SQUARE = (1024, 1024)
    LARGE_PORTRAIT = (1024, 1536)
    LARGE_LANDSCAPE = (1536, 1024)
    LARGE_SQUARE = (1472, 1472)
    WALLPAPER_PORTRAIT = (1088, 1920)
    WALLPAPER_LANDSCAPE = (1920, 1088)


class Sampler(Enum):
    EULER = "k_euler"
    EULER_ANC = "k_euler_ancestral"
    DPM = "k_dpmpp_2s_ancestral"


class Noise(Enum):
    NATIVE = "native"
    KARRAS = "karras"
    EXPONENTIAL = "exponential"
    POLYEXPONENTIAL = "polyexponential"



import os
import base64
import asyncio
import unittest
from unittest.mock import AsyncMock, MagicMock

from httpx import AsyncClient

from novelai import NAIClient, Metadata, Host, Model, Action, Resolution
from novelai.exceptions import (
    APIError,
    AuthError,
    ConcurrentError,
    TimeoutError,
    NovelAIError,
)


with open("tests/images/portrait.jpg", "rb") as f:
    base_image = base64.b64encode(f.read()).decode("utf-8")

with open("tests/images/inpaint_left.jpg", "rb") as f:
    mask = base64.b64encode(f.read()).decode("utf-8")


class TestGenerateImage(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.naiclient = NAIClient(
            os.getenv("USERNAME") or "test_username",
            os.getenv("PASSWORD") or "test_password",
        )

    @unittest.skipIf(
        not (os.getenv("USERNAME") and os.getenv("PASSWORD")),
        "Skipped due to missing environment variables",
    )
    async def test_generate(self):
        await self.naiclient.init()
        metadata = Metadata(prompt="1girl")

        async def task_api():
            with self.subTest("task_api"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.API
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="generate_api.png", verbose=True)
                        self.assertTrue(image.filename == "generate_api.png")
                except ConcurrentError:
                    self.skipTest("task_api raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_api timed out")

        async def task_web():
            with self.subTest("task_web"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.WEB
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="generate_web.png", verbose=True)
                        self.assertTrue(image.filename == "generate_web.png")
                except ConcurrentError:
                    self.skipTest("task_web raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_web timed out")

        tasks = [
            asyncio.create_task(task_api()),
            asyncio.create_task(task_web()),
        ]
        await asyncio.wait(tasks)

    @unittest.skipIf(
        not (os.getenv("USERNAME") and os.getenv("PASSWORD")),
        "Skipped due to missing environment variables",
    )
    async def test_img2img(self):
        await self.naiclient.init()
        metadata = Metadata(
            prompt="1girl",
            action=Action.IMG2IMG,
            res_preset=Resolution.NORMAL_PORTRAIT,
            image=base_image,
            strength=0.45,
            noise=0.1,
        )

        async def task_api():
            with self.subTest("task_api"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.API
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="img2img_api.png", verbose=True)
                except ConcurrentError:
                    self.skipTest("task_api raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_api timed out")

        async def task_web():
            with self.subTest("task_web"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.WEB
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="img2img_web.png", verbose=True)
                except ConcurrentError:
                    self.skipTest("task_web raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_web timed out")

        tasks = [
            asyncio.create_task(task_api()),
            asyncio.create_task(task_web()),
        ]
        await asyncio.wait(tasks)

    @unittest.skipIf(
        not (os.getenv("USERNAME") and os.getenv("PASSWORD")),
        "Skipped due to missing environment variables",
    )
    async def test_inpaint(self):
        await self.naiclient.init()
        metadata = Metadata(
            prompt="1girl",
            model=Model.V3INP,
            action=Action.INPAINT,
            res_preset=Resolution.NORMAL_PORTRAIT,
            image=base_image,
            mask=mask,
        )

        async def task_api():
            with self.subTest("task_api"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.API
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="inpaint_api.png", verbose=True)
                except ConcurrentError:
                    self.skipTest("task_api raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_api timed out")

        async def task_web():
            with self.subTest("task_web"):
                try:
                    output = await self.naiclient.generate_image(
                        metadata, host=Host.WEB
                    )
                    self.assertTrue(len(output) > 0)
                    for image in output:
                        image.save(filename="inpaint_web.png", verbose=True)
                except ConcurrentError:
                    self.skipTest("task_web raised concurrent error")
                except TimeoutError:
                    self.skipTest("task_web timed out")

        tasks = [
            asyncio.create_task(task_api()),
            asyncio.create_task(task_web()),
        ]
        await asyncio.wait(tasks)

    def test_exceptions(self):
        self.naiclient.client = AsyncClient()
        metadata = Metadata(prompt="1girl", seed=1, extra_noise_seed=2)
        error_codes = [400, 401, 402, 409, 429, 500]
        error_exceptions = [
            APIError,
            AuthError,
            AuthError,
            NovelAIError,
            ConcurrentError,
            NovelAIError,
        ]

        for code, exception in zip(error_codes, error_exceptions):
            with self.subTest(f"Status code: {code}"):
                self.naiclient.running = True
                # Mock the AsyncClient's post method to return an error code
                self.naiclient.client.post = AsyncMock(
                    return_value=MagicMock(status_code=code)
                )

                # Check if correct exception is raised
                with self.assertRaises(exception):
                    asyncio.run(self.naiclient.generate_image(metadata))


if __name__ == "__main__":
    unittest.main()



class AuthError(Exception):
    """
    Exception for account authentication errors.
    """

    pass


class APIError(Exception):
    """
    Exception for package-level errors which need to be fixed in the future development (e.g. validation errors).
    """

    pass


class NovelAIError(Exception):
    """
    Exception for errors returned from NovelAI which are not handled by the package.
    """

    pass


class TimeoutError(NovelAIError):
    """
    Exception for request timeouts.
    """

    pass


class ConcurrentError(NovelAIError):
    """
    Exception for concurrent request errors.
    """

    pass



import io
import zipfile
from typing import Generator
from base64 import urlsafe_b64encode
from hashlib import blake2b

import argon2

from .types import User
from .exceptions import APIError, AuthError, NovelAIError, ConcurrentError


# https://github.com/Aedial/novelai-api/blob/main/novelai_api/utils.py
def encode_access_key(user: User) -> str:
    """
    Generate hashed access key from the user's username and password using the blake2 and argon2 algorithms.

    Parameters
    ----------
    user : `novelai.types.User`
        User object containing username and password

    Returns
    -------
    `str`
        Hashed access key
    """
    pre_salt = f"{user.password[:6]}{user.username}novelai_data_access_key"

    blake = blake2b(digest_size=16)
    blake.update(pre_salt.encode())
    salt = blake.digest()

    raw = argon2.low_level.hash_secret_raw(
        secret=user.password.encode(),
        salt=salt,
        time_cost=2,
        memory_cost=int(2000000 / 1024),
        parallelism=1,
        hash_len=64,
        type=argon2.low_level.Type.ID,
    )
    hashed = urlsafe_b64encode(raw).decode()

    return hashed[:64]


class ResponseParser:
    """
    A helper class to parse the response from NovelAI's API.

    Parameters
    ----------
    response : `httpx.Response`
        Response object from the API
    """

    def __init__(self, response):
        self.response = response

    def handle_status_code(self):
        """
        Handle the status code of the response.

        Raises
        ------
        `novelai.exceptions.APIError`
            If the status code is 400
        `novelai.exceptions.AuthError`
            If the status code is 401 or 402
        `novelai.exceptions.ConcurrentError`
            If the status code is 429
        `novelai.exceptions.NovelAIError`
            If the status code is 409 or any other unknown status code
        """
        match self.response.status_code:
            case 200 | 201:
                return
            case 400:
                raise APIError(
                    f"A validation error occured. Message from NovelAI: {self.response.json().get('message')}"
                )
            case 401:
                self.running = False
                raise AuthError(
                    f"Access token is incorrect. Message from NovelAI: {self.response.json().get('message')}"
                )
            case 402:
                self.running = False
                raise AuthError(
                    f"An active subscription is required to access this endpoint. Message from NovelAI: {self.response.json().get('message')}"
                )
            case 409:
                raise NovelAIError(
                    f"A conflict error occured. Message from NovelAI: {self.response.json().get('message')}"
                )
            case 429:
                raise ConcurrentError(
                    f"A concurrent error occured. Message from NovelAI: {self.response.json().get('message')}"
                )
            case _:
                raise NovelAIError(
                    f"An unknown error occured. Error message: {self.response.status_code} {self.response.reason_phrase}"
                )

    def parse_zip_content(self) -> Generator[bytes, None, None]:
        """
        Parse binary data of a zip file into a dictionary.

        Parameters
        ----------
        zip_data : `bytes`
            Binary data of a zip file

        Returns
        -------
        `Generator`
            A generator of binary data of all files in the zip
        """
        with zipfile.ZipFile(io.BytesIO(self.response.content)) as zip_file:
            for filename in zip_file.namelist():
                yield zip_file.read(filename)



from pydantic.dataclasses import dataclass


@dataclass(frozen=True, kw_only=True, slots=True)
class HostInstance:
    """
    Hostnames of NovelAI services and corresponding accepted content types.
    """

    url: str
    accept: str
